@@ -14,9 +14,7 @@ def exists(val):
 def default(val, d):
     return val if exists(val) else d
 
-# triton - fused feedforward (wip)
-
-# todo - modify to be bmm
+# triton
 
 @triton.autotune(
     configs=[
@@ -80,10 +78,6 @@ def bmm_kernel(
     o_ptrs = o_ptr + stride_om * offs_m[:, None] + stride_on * offs_n[None, :] + stride_ol * pid_batch
     tl.store(o_ptrs, o, mask=mask)
 
-@triton.jit
-def relu_squared_activation(x):
-    return tl.where(x > 0, x * x, 0.)
-
 def triton_bmm(x, y, activation = None):
     B, M, K = x.shape
 
@@ -109,10 +103,114 @@ def triton_bmm(x, y, activation = None):
     )
     return o
 
+@triton.autotune(
+    configs=[
+        triton.Config({'BLOCK_SIZE_M': 128, 'BLOCK_SIZE_N': 128, 'BLOCK_SIZE_K': 32, 'GROUP_SIZE_M': 8}, num_stages=4, num_warps=4),
+        triton.Config({'BLOCK_SIZE_M': 128, 'BLOCK_SIZE_N': 64 , 'BLOCK_SIZE_K': 32, 'GROUP_SIZE_M': 8}, num_stages=4, num_warps=4),
+        triton.Config({'BLOCK_SIZE_M': 64 , 'BLOCK_SIZE_N': 128, 'BLOCK_SIZE_K': 32, 'GROUP_SIZE_M': 8}, num_stages=4, num_warps=4),
+        triton.Config({'BLOCK_SIZE_M': 128, 'BLOCK_SIZE_N': 32 , 'BLOCK_SIZE_K': 32, 'GROUP_SIZE_M': 8}, num_stages=4, num_warps=4),
+        triton.Config({'BLOCK_SIZE_M': 64 , 'BLOCK_SIZE_N': 32 , 'BLOCK_SIZE_K': 32, 'GROUP_SIZE_M': 8}, num_stages=5, num_warps=2),
+        triton.Config({'BLOCK_SIZE_M': 32 , 'BLOCK_SIZE_N': 64 , 'BLOCK_SIZE_K': 32, 'GROUP_SIZE_M': 8}, num_stages=5, num_warps=2),
+    ],
+    key=['M', 'N', 'K'],
+)
+@triton.jit
+def bmm_with_bias_kernel(
+    x_ptr, w_ptr, b_ptr, o_ptr,
+    M, N, K,
+    stride_xl, stride_xm, stride_xk,
+    stride_wl, stride_wk, stride_wn,
+    stride_bl, stride_bm, stride_bn,
+    stride_ol, stride_om, stride_on,
+    **meta,
+):
+    BLOCK_SIZE_M = meta['BLOCK_SIZE_M']
+    BLOCK_SIZE_N = meta['BLOCK_SIZE_N']
+    BLOCK_SIZE_K = meta['BLOCK_SIZE_K']
+    GROUP_SIZE_M = 8
+
+    pid_batch = tl.program_id(0)
+    pid = tl.program_id(1)
+
+    num_pid_m = tl.cdiv(M, BLOCK_SIZE_M)
+    num_pid_n = tl.cdiv(N, BLOCK_SIZE_N)
+    num_pid_in_group = GROUP_SIZE_M * num_pid_n
+    group_id = pid // num_pid_in_group
+    first_pid_m = group_id * GROUP_SIZE_M
+    group_size_m = min(num_pid_m - first_pid_m, GROUP_SIZE_M)
+    pid_m = first_pid_m + (pid % group_size_m)
+    pid_n = (pid % num_pid_in_group) // group_size_m
+
+    offs_m = pid_m * BLOCK_SIZE_M + tl.arange(0, BLOCK_SIZE_M)
+    offs_n = pid_n * BLOCK_SIZE_N + tl.arange(0, BLOCK_SIZE_N)
+
+    offs_k = tl.arange(0, BLOCK_SIZE_K)
+
+    x_ptrs = x_ptr + (offs_m[:, None]*stride_xm + offs_k [None, :]*stride_xk + pid_batch*stride_xl)
+    w_ptrs = w_ptr + (offs_k [:, None]*stride_wk + offs_n[None, :]*stride_wn + pid_batch*stride_wl)
+    b_ptrs = b_ptr + (offs_m[:, None]*stride_bm + offs_n[None, :]*stride_bn + pid_batch*stride_bl)
+
+    o = tl.zeros((BLOCK_SIZE_M, BLOCK_SIZE_N), dtype=tl.float32)
+    for k in range(0, K, BLOCK_SIZE_K):
+        x = tl.load(x_ptrs)
+        w = tl.load(w_ptrs)
+        o += tl.dot(x, w)
+
+        x_ptrs += BLOCK_SIZE_K * stride_xk
+        w_ptrs += BLOCK_SIZE_K * stride_wk
+
+    biases = tl.load(b_ptrs)
+    o = o + biases
+
+    if exists(meta['ACTIVATION']):
+        o = meta['ACTIVATION'](o)
+
+    offs_m = pid_m * BLOCK_SIZE_M + tl.arange(0, BLOCK_SIZE_M)
+    offs_n = pid_n * BLOCK_SIZE_N + tl.arange(0, BLOCK_SIZE_N)
+    mask = (offs_m[:, None] < M) & (offs_n[None, :] < N)
+
+    o_ptrs = o_ptr + stride_om * offs_m[:, None] + stride_on * offs_n[None, :] + stride_ol * pid_batch
+    tl.store(o_ptrs, o, mask=mask)
+
+def triton_bmm_with_bias(x, w, b, activation = None):
+    B, M, K = x.shape
+
+    if w.ndim == 2:
+        w = w.unsqueeze(0).expand(B, -1, -1)
+
+    _, K, N = w.shape
+
+    if b.ndim == 1:
+        b = b.unsqueeze(0).unsqueeze(1).expand(B, M, -1)
+
+    assert (K % 32 == 0), "K must be divisible by 32"
+
+    o = torch.empty((B, M, N), device = x.device, dtype = x.dtype)
+
+    grid = lambda META: (
+        B, triton.cdiv(M, META['BLOCK_SIZE_M']) * triton.cdiv(N, META['BLOCK_SIZE_N']),
+    )
+
+    bmm_with_bias_kernel[grid](
+        x, w, b, o,
+        M, N, K,
+        x.stride(0), x.stride(1), x.stride(2),
+        w.stride(0), w.stride(1), w.stride(2),
+        b.stride(0), b.stride(1), b.stride(2),
+        o.stride(0), o.stride(1), o.stride(2),
+        ACTIVATION = activation
+    )
+
+    return o
+
+@triton.jit
+def relu_squared_activation(x):
+    return tl.where(x > 0, x * x, 0.)
+
 class _relu_squared(autograd.Function):
     @classmethod
-    def forward(self, ctx, x, w):
-        o = triton_bmm(x, w, activation = relu_squared_activation)
+    def forward(self, ctx, x, w, b):
+        o = triton_bmm_with_bias(x, w, b, activation = relu_squared_activation)
         ctx.save_for_backward(x, w, o)
         return o
 
@@ -120,20 +218,20 @@ class _relu_squared(autograd.Function):
     def backward(self, ctx, dy):
         x, w, o = ctx.saved_tensors
         dy = torch.sqrt(o) * 2 * dy
+        db = dy.sum(dim = (0, 1))
         dx = triton_bmm(dy, w.t())
         dw = triton_bmm(x.transpose(-1, -2), dy)
-        return dx, dw
+        return dx, dw, db
 
 triton_relu_squared = _relu_squared.apply
 
-def fused_relu_squared(x, w, use_triton = False):
+def fused_relu_squared(x, w, b, use_triton = False):
     if use_triton:
-        return triton_relu_squared(x, w)
-    else:
-        proj = x @ w
-        return F.relu(proj) ** 2
+        return triton_relu_squared(x, w, b)
 
-# triton - softmax (wip)
+    return F.relu(x @ w + b) ** 2
+
+# triton - softmax
 
 def calc_num_warps(block_size):
     num_warps = 4
@@ -282,7 +380,6 @@ def cross_entropy_fn(logits, labels, ignore_index = 0., use_triton = False):
 @triton.jit
 def layernorm_kernel_forward_training(
     output_ptr,
-    normed_ptr,
     mean_centered_ptr,
     inv_var_ptr,
     input_ptr,
@@ -292,7 +389,6 @@ def layernorm_kernel_forward_training(
     gamma_row_stride,
     beta_row_stride,
     output_row_stride,
-    normed_row_stride,
     mean_centered_row_stride,
     inv_var_row_stride,
     n_cols,
@@ -328,10 +424,6 @@ def layernorm_kernel_forward_training(
     output_ptrs = output_row_start_ptr + col_offsets
     tl.store(output_ptrs, output, mask=mask)
 
-    normed_row_start_ptr = normed_ptr + row_idx * normed_row_stride
-    normed_ptrs = normed_row_start_ptr + col_offsets
-    tl.store(normed_ptrs, normed, mask=mask)
-
     mean_centered_row_start_ptr = mean_centered_ptr + row_idx * mean_centered_row_stride
     mean_centered_ptrs = mean_centered_row_start_ptr + col_offsets
     tl.store(mean_centered_ptrs, row_mean_centered, mask=mask)
@@ -387,12 +479,10 @@ def layernorm_kernel_forward_inference(
 def layernorm_kernel_backward(
     output_ptr,
     dy_ptr,
-    normed_ptr,
     mean_centered_ptr,
     inv_var_ptr,
     output_row_stride,
     dy_row_stride,
-    normed_row_stride,
     mean_centered_row_stride,
     inv_var_row_stride,
     n_cols,
@@ -402,23 +492,21 @@ def layernorm_kernel_backward(
     BLOCK_SIZE = meta['BLOCK_SIZE']
 
     dy_row_start_ptr = dy_ptr + row_idx * dy_row_stride
-    normed_row_start_ptr = normed_ptr + row_idx * normed_row_stride
     mean_centered_row_start_ptr = mean_centered_ptr + row_idx * mean_centered_row_stride
     inv_var_row_start_ptr = inv_var_ptr + row_idx * inv_var_row_stride
 
     col_offsets = tl.arange(0, BLOCK_SIZE)
     dy_ptrs = dy_row_start_ptr + col_offsets
-    normed_ptrs = normed_row_start_ptr + col_offsets
     mean_centered_ptrs = mean_centered_row_start_ptr + col_offsets
     inv_var_ptrs = inv_var_row_start_ptr + col_offsets
 
     mask = col_offsets < n_cols
 
     dy = tl.load(dy_ptrs, mask=mask, other=0.)
-    normed = tl.load(normed_ptrs, mask=mask, other=0.)
     mean_centered = tl.load(mean_centered_ptrs, mask=mask, other=0.)
     inv_var = tl.load(inv_var_ptrs, mask=mask, other=0.)
 
+    normed = mean_centered * inv_var
     output = 1. / n_cols * inv_var * (n_cols * dy - tl.sum(dy, axis = 0) - normed * tl.sum(dy * normed, axis = 0))
 
     output_row_start_ptr = output_ptr + row_idx * output_row_stride
@@ -442,13 +530,11 @@ class _layernorm(autograd.Function):
         out = torch.empty_like(x)
 
         if training:
-            normed_x = torch.empty_like(x)
             scaled_x = torch.empty_like(x)
             inv_var = torch.empty_like(x)
 
             layernorm_kernel_forward_training[(n_rows,)](
                 out,
-                normed_x,
                 scaled_x,
                 inv_var,
                 x,
@@ -458,7 +544,6 @@ class _layernorm(autograd.Function):
                 expanded_gamma.stride(0),
                 expanded_beta.stride(0),
                 out.stride(0),
-                normed_x.stride(0),
                 scaled_x.stride(0),
                 inv_var.stride(0),
                 n_cols,
@@ -466,7 +551,7 @@ class _layernorm(autograd.Function):
                 num_warps = num_warps,
                 BLOCK_SIZE = BLOCK_SIZE,
             )
-            ctx.save_for_backward(scaled_x, normed_x, gamma, inv_var)
+            ctx.save_for_backward(scaled_x, gamma, inv_var)
         else:
             layernorm_kernel_forward_inference[(n_rows,)](
                 out,
@@ -491,13 +576,13 @@ class _layernorm(autograd.Function):
         dim = shape[-1]
         dy = dy.view(-1, dim)
 
-        scaled_x, normed_x, gamma, inv_var = ctx.saved_tensors
+        scaled_x, gamma, inv_var = ctx.saved_tensors
 
         dbeta = dy.sum(dim = 0)
-        dgamma = (dy * normed_x).sum(dim = 0)
+        dgamma = (dy * scaled_x * inv_var).sum(dim = 0)
         dxhat = dy * gamma
 
-        n_rows, n_cols = normed_x.shape
+        n_rows, n_cols = dy.shape
 
         BLOCK_SIZE = triton.next_power_of_2(n_cols)
         num_warps = calc_num_warps(BLOCK_SIZE)
@@ -507,12 +592,10 @@ class _layernorm(autograd.Function):
         layernorm_kernel_backward[(n_rows,)](
             dx,
             dxhat,
-            normed_x,
             scaled_x,
             inv_var,
             dx.stride(0),
             dxhat.stride(0),
-            normed_x.stride(0),
             scaled_x.stride(0),
             inv_var.stride(0),
             n_cols,
@@ -585,6 +668,7 @@ class FeedForward(nn.Module):
         self.norm = nn.LayerNorm(dim)
 
         self.proj_in_weight = nn.Parameter(torch.randn(dim, inner_dim))
+        self.proj_in_bias = nn.Parameter(torch.randn(inner_dim))
         self.proj_out = nn.Linear(inner_dim, dim)
 
     def forward(self, x, use_triton = None):
@@ -592,7 +676,7 @@ class FeedForward(nn.Module):
 
         x = layernorm(x, self.norm.weight, self.norm.bias, use_triton = False, training = self.training)
 
-        x = fused_relu_squared(x, self.proj_in_weight, use_triton = use_triton)
+        x = fused_relu_squared(x, self.proj_in_weight, self.proj_in_bias, use_triton = use_triton)
         x = self.proj_out(x)
         return x
 
